// GRAPHICS.CPP
// Mode 13h Graphics Library
//
// DOS 16-bit Real Mode
// Tested with Borland C++ 5.0
//
// Make sure that this library is compiled and linked with the htimer
// library.

#include <dos.h>
#include <alloc.h>
#include <mem.h>
#include <stdio.h>
#include "graphics.h"
#include "types.h"
#include "htimer.h"

// VGA palette ports

#define PALETTE_MASK      0x3C6
#define PALETTE_REG_RD    0x3C7
#define PALETTE_REG_WR    0x3C8
#define PALETTE_DATA      0x3C9

#define VGA_INPUT_STATUS  0x3DA
#define VGA_VSYNC_MASK    0x08

// Pointer to ROM character set
byte far *screen     = (byte far *)0xA0000000L;
byte far *curBuf     = screen;
byte far *ROMCharSet = (byte far *)0xF000FA6EL;


// Screen Functions ////////////////////////////////////////////////////////////

void setvmode(int mode)                // Uses BIOS INT 10h to set video mode
{
  asm mov ax, mode
  asm int 0x10
}

int loadScreen(char *file, byte far *buffer)
// Loads a 64K file into buffer
{
  FILE *handle;                                 // Opens file and tests for
  handle = fopen(file, "rb");                   // errors.  If file does not
  if (handle == NULL)                           // exist, function returns 1.
    return 1;

  unsigned index;                               // Loads data
  for (index = 0; index < 64000U; index++)
    buffer[index] = (byte)fgetc(handle);

  fclose(handle);                               // Normal exit
  return 0;
}

int saveScreen(char *file, byte far *buffer)
// Saves a 64K buffer into file
{
  FILE *handle;
  handle = fopen(file, "wb");
  if (handle == NULL)
    return 1;

  unsigned index;
  for (index = 0; index < 64000U; index++)
    fputc(buffer[index], handle);

  fclose(handle);
  return 0;
}

void clearScreen(byte color)
{
  unsigned scrseg = FP_SEG(screen),
           scroff = FP_OFF(screen);

  asm
  {
    push es
    push di

    mov cx, 32000
    mov al, color
    mov ah, al

    mov es, scrseg
    mov di, scroff

    rep stosw

    pop di
    pop es
  }
}

void blitScreen(byte far *buffer)               // Writes buffer to screen
{
  unsigned bufseg = FP_SEG(buffer),             // Gets segments and offsets
           bufoff = FP_OFF(buffer),             // of screen and buffer

           scrseg = FP_SEG(screen),
           scroff = FP_OFF(screen);

  asm
  {
    push ds                                    // PUSH so we don't lose
    push si                                    // segment registers
    push es
    push di

    mov ds, bufseg                             // Set DS:SI to buffer
    mov si, bufoff

    mov es, scrseg                             // Set ES:DI to screen
    mov di, scroff

    mov cx, 32000                           // Number of pixels to be copied / 2
    rep movsw                               // Copy pixels (faster with words)

    pop di                                     // Restore segment registers
    pop es
    pop si
    pop ds
  }
}


// Palette Functions ///////////////////////////////////////////////////////////

void setPalette(int x, colorType *color) // Sets one palette register to a
{                                            // value
  outp(PALETTE_MASK, 0xFF);     // Sets VGA Palette Mask to include all ports
  outp(PALETTE_REG_WR, x);       // Tells VGA which color we are working with
  outp(PALETTE_DATA, color->red);           // OUTs red
  outp(PALETTE_DATA, color->green);         // OUTs green
  outp(PALETTE_DATA, color->blue);          // OUTs blue
}

void getPalette(int x, colorType *color) // Reads one palette register
{                                        // value
  outp(PALETTE_MASK, 0xFF);     // Sets VGA Palette Mask to include all ports
  outp(PALETTE_REG_RD, x);      // Tells VGA which color we are working with
  color->red   = inp(PALETTE_DATA);   // INs red
  color->green = inp(PALETTE_DATA);   // INs green
  color->blue  = inp(PALETTE_DATA);   // INs blue
}

void setAllPalette(paletteType palette)  // Sets all palette registers
{
  int x;
  for (x = 0; x < 256; x++)               // Cycles through registers and...
    setPalette(x, &palette[x]);           // Sets colors
}

void getAllPalette(paletteType palette)  // Gets all palette registers
{
  int x;
  for (x = 0; x < 256; x++)                // Cycles through registers and...
    getPalette(x, &palette[x]);            // Gets colors
}

int loadPalette(char *file, paletteType palette)  // Loads palette from file
{
  FILE *handle;
  handle = fopen(file, "rb");               // Sees if file exists.  If not,
  if (handle == NULL)                       // returns with value 1.
    return 1;

  fread(palette, sizeof(colorType) * 256, 1, handle);

  fclose(handle);                           // Normal exit
  return 0;
}

int savePalette(char *file, paletteType palette)
{
  FILE *handle;
  handle = fopen(file, "wb");
  if (handle == NULL)
    return 1;

  fwrite(palette, sizeof(colorType) * 256, 1, handle);

  fclose(handle);
  return 0;
}

void resetPalette(void)                 // Sets all palette registers to black
{
  int x;                                // index
  colorType empty = {0, 0, 0};          // Black color value
  for (x = 0; x < 256; x++)             // Cycles through registers and...
    setPalette(x, &empty);                // Sets to black
}

paletteType tempPal;

void fadeOut(paletteType from, long msecs)
{
  HTimer ht;

  int ic, iv;
  for (ic = 15; ic > -1; ic--)
  {
    for (iv = 0; iv < 256; iv++)
    {
      tempPal[iv].red   = (float)from[iv].red   * (float)ic / (float)15;
      tempPal[iv].green = (float)from[iv].green * (float)ic / (float)15;
      tempPal[iv].blue  = (float)from[iv].blue  * (float)ic / (float)15;
    }
    setAllPalette(tempPal);

    ht.timerOn();
    while (ht.getElapsed() < msecs) {}
    ht.timerOff();
  }
}

void fadeIn(paletteType to, long msecs)
{
  HTimer ht;

  int ic, iv;
  for (ic = 0; ic < 16; ic++)
  {
    for (iv = 0; iv < 256; iv++)
    {
      tempPal[iv].red   = (float)to[iv].red   * (float)ic / (float)15;
      tempPal[iv].green = (float)to[iv].green * (float)ic / (float)15;
      tempPal[iv].blue  = (float)to[iv].blue  * (float)ic / (float)15;
    }
    setAllPalette(tempPal);

    ht.timerOn();
    while (ht.getElapsed() < msecs) {}
    ht.timerOff();
  }
}

void fadeTo(paletteType from, paletteType to, long msecs)
{
  HTimer ht;

  int ic, iv;
  for (ic = 0; ic < 16; ic++)
  {
    for (iv = 0; iv < 256; iv++)
    {
      tempPal[iv].red   = ((float)from[iv].red *   (15.0 - ic) +
                           (float)to[iv].red   * ic) / 15.0;

      tempPal[iv].green = ((float)from[iv].green * (15.0 - ic) +
                           (float)to[iv].green * ic) / 15.0;

      tempPal[iv].blue  = ((float)from[iv].blue *  (15.0 - ic) +
                           (float)to[iv].blue  * ic) / 15.0;
    }
    setAllPalette(tempPal);

    ht.timerOn();
    while (ht.getElapsed() < msecs) {}
    ht.timerOff();
  }
}


// Image Functions /////////////////////////////////////////////////////////////

void allocImage(int x, int y, imageType *image)  // Initializes and allocates
{                                                // memory for image
  image->x = x;
  image->y = y;
  image->ptr = (byte far *)farmalloc((long)(x * y));
}

void resetImage(imageType *image)              // Deinits image and frees mem
{
  image->x =
  image->y = 0;
  farfree(image->ptr);
}

int loadImage(char *file, imageType *image)
{
  FILE *handle;
  handle = fopen(file, "rb");
  if (handle == NULL)
    return 1;

  long x;
  for (x = 0; x < (long)image->x * (long)image->y; x++)
    image->ptr[x] = fgetc(handle);

  fclose(handle);
  return 0;
}

int saveImage(char *file, imageType *image)
{
  FILE *handle;
  handle = fopen(file, "wb");
  if (handle == NULL)
    return 1;

  long x;
  for (x = 0; x < (long)image->x * (long)image->y; x++)
    fputc(image->ptr[x], handle);

  fclose(handle);
  return 0;
}

void grabImage(int x, int y, imageType *image, byte far *buffer)
{
  int ix, iy;
  for (iy = 0; iy < image->y; iy++)
    for (ix = 0; ix < image->x; ix++)
      image->ptr[iy * image->y + ix] = buffer[(iy + y) * 320 + (ix + x)];
}

void blitImageOpaque(int x, int y, imageType *image)
{
  int i;                               // index
  for (i = 0; i < image->y; i++)    // Loops through lines
    _fmemcpy(curBuf + ((i + y) << 8) + ((i + y) << 6) + x,
             image->ptr + i * image->x,
             image->x);
}

void blitImageTransparent(int x, int y, imageType *image)
{
  int ix, iy;
  for (iy = 0; iy < image->y; iy++)
    for (ix = 0; ix < image->x; ix++)
      if (image->ptr[(iy + y) * image->x + ix])
      
        curBuf[((y + iy) << 8) + ((y + iy) << 6) + x + ix] =
        image->ptr[iy * image->x + ix];  
}

void clearArea(int x1, int y1, int x2, int y2, int color)
{
  int i;
  for (i = y1; i < y1 + y2; i++)   // Cycles through rows and clears row
    _fmemset(curBuf + (i << 8) + (i << 6) + x1, color, x2);
}


// Text Functions //////////////////////////////////////////////////////////////

void blitChar(int xc, int yc, int color, char c)
{
  int offset, x, y;
  byte far *workChar;
  byte bitMask = 0x80;

  workChar = ROMCharSet + (c << 3);

  offset = (yc << 8) + (yc << 6) + xc;

  for (y = 0; y < 8; y++)
  {
    bitMask = 0x80;
    for (x = 0; x < 8; x++)
    {
      if (*workChar & bitMask)
        curBuf[offset + x] = color;

      bitMask >>= 1;
    }

    offset += 320;
    workChar++;
  }
}

void blitString(int x, int y, int color, char *string)
{
  int index;
  for (index = 0; string[index]; index++)
    blitChar(x + (index << 3), y, color, string[index]);
}

// Miscellaneous /////////////////////////////////////////////////////////////

void vsync(void)
{
  while (inp(VGA_INPUT_STATUS) & VGA_VSYNC_MASK) {}
  while (!(inp(VGA_INPUT_STATUS) & VGA_VSYNC_MASK)) {}
}

