// GRAPHICS.H
// Graphics Library Header

#ifndef __GRAPHICS_H
#define __GRAPHICS_H

#define byte unsigned char

#define pixel(x, y) (screen[((y) << 8) + ((y) << 6) + (x)])

struct colorType                             // RGB color type for VGA cards
{
  byte red;
  byte green;
  byte blue;
};

struct imageType                             // Type for images of any size
{
  int x, y;                 // X and Y dimensions
  byte far *ptr;            // Pointer to image
};

typedef colorType paletteType[256];

extern byte far *screen;
extern byte far *curBuf;

extern void setvmode(int mode);
extern int loadScreen(char *file, byte far *buffer = screen);
extern int saveScreen(char *file, byte far *buffer = screen);
extern void clearScreen(byte color = 0);
extern void blitScreen(byte far *buffer);

extern void setPalette(int x, colorType *color);
extern void getPalette(int x, colorType *color);
extern void setAllPalette(paletteType palette);
extern void getAllPalette(paletteType palette);
extern int loadPalette(char *file, paletteType palette);
extern int savePalette(char *file, paletteType palette);
extern void resetPalette(void);
extern void fadeOut(paletteType from, long msecs = 10000);
extern void fadeIn(paletteType to, long msecs = 10000);
extern void fadeTo(paletteType from, paletteType to, long msecs = 10000);

extern void allocImage(int x, int y, imageType *image);
extern void resetImage(imageType *image);
extern int loadImage(char *file, imageType *image);
extern int saveImage(char *file, imageType *image);
extern void grabImage(int x, int y, imageType *image,
                      byte far *buffer = screen);
#define blitImage(x, y, z) blitImageOpaque(x, y, z)
extern void blitImageOpaque(int x, int y, imageType *image);
extern void blitImageTransparent(int x, int y, imageType *image);

extern void clearArea(int x1, int y1, int x2, int y2, int color = 0);
extern void blitChar(int xc, int yc, int color, char c);
extern void blitString(int x, int y, int color, char *string);

extern void vsync(void);

#endif
