// Includes
#include <dos.h>
#include <conio.h>
#include "keyboard.h"

// Keyboard ports
#define KEY_INT     0x09
#define KEY_BUFFER  0x60
#define KEY_CONTROL 0x61
#define INT_CONTROL 0x20

void interrupt (far *oldKeyHandler)();    // Old key handler variable

volatile byte rawKey;                      // Most recent scancode
volatile byte keyTable[128];               // Table of keys currently pressed

byte newInput;                  // 1 if newKeyHandler(), 0 if oldKeyHandler()

void interrupt far newKeyHandler()        // New key handler vector
{
  asm
  {
    cli                            // Disables interrupts
    in al, KEY_BUFFER              // Reads value in key buffer
    mov rawKey, al                 // Moves scancode to rawKey
    in al, KEY_CONTROL             // Reads control register from KEY_CONTROL
    or al, 0x82                    // ORs value with 82h
    out KEY_CONTROL, al            // Writes value back to port
    and al, 0x7F                   // ANDs value with 7Fh
    out KEY_CONTROL, al            // Writes value to port again
    mov al, 0x20                   // Write 20h to port 20h to...
    out INT_CONTROL, al            // reset 8259 interrupt controller
  }
  
  if (rawKey < 128)                // If code is a make...
    keyTable[rawKey] = 1;            // Turn on key
  else                             // If code is a break...
    keyTable[rawKey - 128] = 0;      // Turn off key
    
  asm sti                          // Enables interrupts
}

void initKeyboard(void)       // Sets up new key handler using DOS.H routines
{
  oldKeyHandler = getvect(KEY_INT);   // Stores old interrupt
  setvect(KEY_INT, newKeyHandler);    // Sets new one
  newInput = 1;         // Informs program that newKeyHandler() is being used
}

void closeKeyboard(void)      // Resets key handler
{
  setvect(KEY_INT, oldKeyHandler);    // Sets interrupt to oldKeyHandler()
  newInput = 0;         // Informs program that oldKeyHandler() is being used
}

byte keyPressed(void)          // Tests to see if a key is pressed.
{                              // Returns 1 if a key is.
  register int x;
  for (x = 0; x < 128; x++)
    if (keyTable[x]) return x;
  return 0;
}

void pause(void)           // Pauses program until user presses a key
{
  if (newInput)                  // If newKeyHandler()...
    while (1)                        // Tests to see if any keys are on
    {
      if (keyPressed())
      {
        while (keyPressed())
        return;
      }
    }
  else                           // If oldKeyHandler()...
    if (getch() == 0) getch();   // Uses standard getch() and tests for
}                                  // special keys
