#include <conio.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include "graphics.h"

#define bool    int
#define true    1
#define false   0
#define byte    unsigned char
#define word    unsigned short
#define dword   unsigned long
#define boolean signed   char

#include "htimer.h"
#include "pcx.h"
#include "keyboard.h"
#include "mouse.h"
extern "C"
{
#include "detect.h"
#include "smix.h"
}

#define INTRO_FILE    "intro.pcx"
#define PONGPIC_FILE  "pong.pcx"
#define SLB_FILE      "pong.slb"
#define CONFIG_FILE   "pong.cfg"

/* Sound Information ********************************************************/
#define XMS_REQUIRED  500
#define SHAREDEMB
#define NUMSOUNDS     7

#define SOUND_INTRO   0
#define SOUND_PONG    1
#define SOUND_BUTTON  2
#define SOUND_BEGIN   3
#define SOUND_HITWALL 4
#define SOUND_LOST    5
#define SOUND_PADDLE  6

#define PlaySound(a) { if (sb_exist && play_sound) start_sound(sounds[a], a, 0); }

char *sound_key[NUMSOUNDS] =
{
  "intro", "pong", "button", "begin", "hitwall", "lost", "paddle"
};

int baseio, irq, dma8, dma16;
SOUND *sounds[NUMSOUNDS];
/****************************************************************************/

void PongInit(void);
void PongClose(void);
void SoundExitProc(void);
void LoadSounds(void);
void FreeSounds(void);
void KMPause(void);
void GetMouseXYButtons(int *x, int *y, int *buttons);
void Intro(void);
void MainMenu(void);
void BeginGame(void);
void DrawScreen(void);
void GenerateGamePal(void);
void HDelay(long msecs);

class BUTTON
{
private:
  int x, y;
  int w, h;
  bool pushed;
  char *message;

public:
  BUTTON(int _x, int _y, int _w, int _h, char *_message)
  {
    x = _x;
    y = _y;
    w = _w;
    h = _h;
    message = _message;
    pushed = false;
  }

  void Draw(int clr, int bc, int dc, int tc);
  void SetPushed(bool _pushed) { pushed = _pushed; }
  int Within(int _x, int _y)
  {
    if (_x >= x && _x <= x + w && _y >= y && _y <= y + h) return 1;
    return 0;
  }
};

void BUTTON::Draw(int clr, int bc, int dc, int tc)
{
  hideMouse();

  byte far *oldbuf = curBuf;
  curBuf = screen;

  clearArea(x, y, w, h, clr);

  int i;

  if (pushed) { int c = bc; bc = dc; dc = c; }  // Swaps color values

  for (i = x - 2; i < x + w + 2; i++)   // Draws horizontal lines
  {
    pixel(i, y - 2)     = bc;
    pixel(i, y + h + 1) = dc;
  }

  for (i = y - 2; i < y + h + 2; i++)   // Draws vertical lines
  {
    pixel(x - 2,     i) = bc;
    pixel(x + w + 1, i) = dc;
  }

  blitString(x, y, tc, message);

  curBuf = oldbuf;
  showMouse();
}

paletteType default_pal;
paletteType game_pal;

int num_players = 1;
int play_sound  = 1;
int game_speed  = 5;
int play_to     = 15;

int sb_exist;

int main(void)
{
  PongInit();
  Intro();
  MainMenu();
  PongClose();

  textcolor(7);
  clrscr();
  cprintf("Pong v1.0\n\r"
          "Chad Austin\n\r"
          "\n"
          "Compile Date: %s\n\r"
          "Compile Time: %s\n\r"
          "\n"
          "Thanks to Ethan Brodsky for providing his freeware sound library.\n\r",
          __DATE__,
          __TIME__);
  if (getch() == 0) getch();

  return 0;
}

void PongInit(void)
{
  randomize();
  FILE *stream = fopen(CONFIG_FILE, "rb");
  if (stream != NULL)
  {
    fread(&num_players, 1, sizeof(num_players), stream);
    fread(&play_sound , 1, sizeof(play_sound ), stream);
    fread(&game_speed , 1, sizeof(game_speed ), stream);
    fread(&play_to    , 1, sizeof(play_to    ), stream);
    fclose(stream);
  }

  int buttons;
  resetMouse(&buttons);
  initKeyboard();
  setvmode(0x13);
  getAllPalette(default_pal);
  LoadSounds();
  if (!detect_settings(&baseio, &irq, &dma8, &dma16)) { sb_exist = 0; return; }
  if (!init_sb(baseio, irq, dma8, dma16)) { sb_exist = 0; return; }
  init_mixing();
  set_sampling_rate(11025);
}

void PongClose(void)
{
  shutdown_mixing();
  shutdown_sb();
  FreeSounds();
  setvmode(0x3);
  closeKeyboard();

  FILE *stream = fopen(CONFIG_FILE, "wb");
  if (stream != NULL)
  {
    fwrite(&num_players, 1, sizeof(num_players), stream);
    fwrite(&play_sound , 1, sizeof(play_sound ), stream);
    fwrite(&game_speed , 1, sizeof(game_speed ), stream);
    fwrite(&play_to    , 1, sizeof(play_to    ), stream);
    fclose(stream);
  }
}

void SoundExitProc(void)
{
  int i;
  for (i = 0; i < NUMSOUNDS; i++)
    if (sounds[i])
      free_sound(sounds + i);

#ifdef SHAREDEMB
  shutdown_sharing();
#endif
}

void LoadSounds(void)
{
  sb_exist = 1;
  if (!init_xms())                 { sb_exist = 0; return; }
  if (getfreexms() < XMS_REQUIRED) { sb_exist = 0; return; }

#ifdef SHAREDEMB
  init_sharing();
#endif

  if (!open_sound_resource_file(SLB_FILE)) { sb_exist = 0; return; }

  int i;
  for (i = 0; i < NUMSOUNDS; i++)
    load_sound(&(sounds[i]), sound_key[i]);

  close_sound_resource_file();
  atexit(SoundExitProc);
}

void FreeSounds(void)
{
  int i;
  for (i = 0; i < NUMSOUNDS; i++)
    free_sound(sounds + i);

#ifdef SHAREDEMB
  shutdown_sharing();
#endif
}

void KMPause(void)
{
  int x, y, buttons = 0;

  while (keyPressed())  {}
  while (!keyPressed())
  {
    GetMouseXYButtons(&x, &y, &buttons);
    if (buttons) return;
  }
}

void GetMouseXYButtons(int *x, int *y, int *buttons)
{
  mouseXYButtons(x, y, buttons);
  *x /= 2;
}

void Intro(void)
{
  PCX_PICTURE intro;

  colorType blank = { 0, 0, 0 };
  for (int i = 0; i < 256; i++)
    setPalette(i, &blank);

  if (PCX_Init(&intro))
  {
    PongClose();
    cputs("Not enough memory\n\r");
    exit(1);
  }

  if (PCX_Load(INTRO_FILE, &intro))
  {
    PCX_Close(&intro);
    PongClose();
    cputs("INTRO.PCX not found\n\r");
    exit(1);
  }

  blitScreen(intro.buffer);
  fadeIn((colorType *)intro.palette);
  PlaySound(SOUND_INTRO);
  KMPause();
  fadeOut((colorType *)intro.palette);

  if (PCX_Load(PONGPIC_FILE, &intro))
  {
    PCX_Close(&intro);
    PongClose();
    cputs("PONG.PCX not found\n\r");
    exit(1);
  }

  blitScreen(intro.buffer);
  fadeIn((colorType *)intro.palette);
  PlaySound(SOUND_PONG);
  KMPause();
  fadeOut((colorType *)intro.palette);

  PCX_Close(&intro);
}

#define STR1 "Number of Players: "
#define STR2 "Speed:             "
#define STR3 "Sound?             "
#define STR4 "Play to:           "
#define COLOR_ARGS 19, 21, 17

#define SPEED_X    60
#define SPEED_Y    30

#define PLAYTO_X   60 + 16
#define PLAYTO_Y   70

void MainMenu(void)
{
  clearScreen(19);

  blitString(10, 10, 2, STR1);
  BUTTON OnePlayer(strlen(STR1) * 8 + 10, 10, 8, 8, "1");
  BUTTON TwoPlayer(strlen(STR1) * 8 + 30, 10, 8, 8, "2");
  if (num_players == 1)
    OnePlayer.SetPushed(true);
  else
    TwoPlayer.SetPushed(true);
  OnePlayer.Draw(COLOR_ARGS, 2);
  TwoPlayer.Draw(COLOR_ARGS, 2);

  BUTTON BNew (10, 170, 40, 8, "Begin");
  BNew.Draw (COLOR_ARGS, 9);

  BUTTON BQuit(60, 170, 32, 8, "Quit");
  BQuit.Draw(COLOR_ARGS, 4);

  blitString(10, 30, 3, STR2);
  BUTTON SpeedUp  (strlen(STR2) * 8 + 10, 30, 8, 8, "");
  BUTTON SpeedDown(strlen(STR2) * 8 + 30, 30, 8, 8, "");
  SpeedUp.Draw  (COLOR_ARGS, 3);
  SpeedDown.Draw(COLOR_ARGS, 3);

  char string[3];
  sprintf(string, "%d", game_speed + 1);
  blitString(SPEED_X, SPEED_Y, 3, string);

  blitString(10, 50, 2, STR3);
  BUTTON SoundOn (strlen(STR3) * 8 + 10, 50, 24, 8, "Yes");
  BUTTON SoundOff(strlen(STR3) * 8 + 50, 50, 16, 8, "No");
  if (play_sound)
    SoundOn.SetPushed(true);
  else
    SoundOff.SetPushed(true);
  SoundOn.Draw(COLOR_ARGS, 2);
  SoundOff.Draw(COLOR_ARGS, 2);

  blitString(10, 70, 3, STR4);
  BUTTON PlaytoUp  (strlen(STR4) * 8 + 10, 70, 8, 8, "");
  BUTTON PlaytoDown(strlen(STR4) * 8 + 30, 70, 8, 8, "");
  sprintf(string, "%d", play_to);
  blitString(PLAYTO_X, PLAYTO_Y, 3, string);
  PlaytoUp.Draw(COLOR_ARGS, 3);
  PlaytoDown.Draw(COLOR_ARGS, 3);

  blitString(10, 90,  14, "Controls:");
  blitString(10, 100, 14, "       Player 1           Player 2");
  blitString(10, 110, 14, "Up:       W                  ");
  blitString(10, 120, 14, "Down:     X                  ");

  showMouse();

  fadeIn(default_pal);

  int x, y, buttons = 0;
  int done = 0;
  while (!done)
  {
    while (buttons)  GetMouseXYButtons(&x, &y, &buttons);
    while (!buttons) GetMouseXYButtons(&x, &y, &buttons);

    if (buttons & 1)
    {
      if (OnePlayer.Within(x, y))
      {
        num_players = 1;
        PlaySound(SOUND_BUTTON);
        OnePlayer.SetPushed(true);     TwoPlayer.SetPushed(false);
        OnePlayer.Draw(COLOR_ARGS, 2); TwoPlayer.Draw(COLOR_ARGS, 2);
      }

      if (TwoPlayer.Within(x, y))
      {
        num_players = 2;
        PlaySound(SOUND_BUTTON);
        OnePlayer.SetPushed(false);    TwoPlayer.SetPushed(true);
        OnePlayer.Draw(COLOR_ARGS, 2); TwoPlayer.Draw(COLOR_ARGS, 2);
      }

      if (SpeedUp.Within(x, y))
      {
        PlaySound(SOUND_BUTTON);
        if (++game_speed == 10) game_speed = 9;
        clearArea(SPEED_X, SPEED_Y, 16, 8, 19);
        sprintf(string, "%d", game_speed + 1);
        blitString(SPEED_X, SPEED_Y, 3, string);
      }

      if (SpeedDown.Within(x, y))
      {
        PlaySound(SOUND_BUTTON);
        if (--game_speed == -1) game_speed = 0;
        clearArea(SPEED_X, SPEED_Y, 16, 8, 19);
        sprintf(string, "%d", game_speed + 1);
        blitString(SPEED_X, SPEED_Y, 3, string);
      }

      if (PlaytoUp.Within(x, y))
      {
        PlaySound(SOUND_BUTTON);
        if (++play_to == 100) play_to = 99;
        clearArea(PLAYTO_X, PLAYTO_Y, 16, 8, 19);
        sprintf(string, "%d", play_to);
        blitString(PLAYTO_X, PLAYTO_Y, 3, string);
      }

      if (PlaytoDown.Within(x, y))
      {
        PlaySound(SOUND_BUTTON);
        if (--play_to == 0) play_to = 1;
        clearArea(PLAYTO_X, PLAYTO_Y, 16, 8, 19);
        sprintf(string, "%d", play_to);
        blitString(PLAYTO_X, PLAYTO_Y, 3, string);
      }

      if (SoundOn.Within(x, y))
      {
        play_sound = 1;
        PlaySound(SOUND_BUTTON);
        SoundOn.SetPushed(true);     SoundOff.SetPushed(false);
        SoundOn.Draw(COLOR_ARGS, 2); SoundOff.Draw(COLOR_ARGS, 2);
      }

      if (SoundOff.Within(x, y))
      {
        play_sound = 0;
        PlaySound(SOUND_BUTTON);
        SoundOn.SetPushed(false);    SoundOff.SetPushed(true);
        SoundOn.Draw(COLOR_ARGS, 2); SoundOff.Draw(COLOR_ARGS, 2);
      }

      if (BNew.Within(x, y))
      {
        PlaySound(SOUND_BEGIN);
        BNew.SetPushed(true);
        BNew.Draw(COLOR_ARGS, 9);
        fadeOut(default_pal);
        hideMouse();

        BeginGame();

        BNew.SetPushed(false);
        BNew.Draw(COLOR_ARGS, 9);
        showMouse();
        fadeIn(default_pal);
      }

      if (BQuit.Within(x, y))
      {
        PlaySound(SOUND_BUTTON);
        BQuit.SetPushed(true);
        BQuit.Draw(COLOR_ARGS, 4);
        done = 1;
      }
    }
  }

  fadeOut(default_pal);
  hideMouse();
}

void BeginGame(void)
{
  GenerateGamePal();

  /* Draw screen */
  saveScreen("tempscrn.tmp");

  DrawScreen();

  fadeIn(game_pal);

#define PADDLE_HEIGHT 32

  imageType paddle1, paddle2, ball;
  allocImage(4, PADDLE_HEIGHT, &paddle1);
  if (paddle1.ptr == NULL)
    { PongClose(); cputs("Error: Not enough memory\n\r"); exit(1); }
  allocImage(4, PADDLE_HEIGHT, &paddle2);
  if (paddle2.ptr == NULL)
    { resetImage(&paddle1); PongClose(); cputs("Error: Not enough memory\n\r"); exit(1); }
  allocImage(3, 3, &ball);
  if (ball.ptr == NULL)
    { resetImage(&paddle1); resetImage(&paddle2); PongClose(); cputs("Error: Not enough memory\n\r"); exit(1); }


  // The first paddle is red.  The second is green.  The ball is white.

  // Generate images
  int ix, iy;

  for (iy = 0; iy < 32; iy++)
    for (ix = 0; ix < 4; ix++)
    {
      if (iy < 16)
      {
        paddle1.ptr[iy * 4 + ix] = (iy % 8) + 40;
        paddle2.ptr[iy * 4 + ix] = (iy % 8) + 48;
      }
      else
      {
        paddle1.ptr[iy * 4 + ix] = ((31 - iy) % 8) + 40;
        paddle2.ptr[iy * 4 + ix] = ((31 - iy) % 8) + 48;
      }
    }

  for (ix = 0; ix < 9; ix++)
    ball.ptr[ix] = 57;

  ball.ptr[1] = ball.ptr[3] = ball.ptr[4] = ball.ptr[5] = ball.ptr[7] = 56;

#define BALL_START_X 320 / 2
#define BALL_START_Y (200 - 32) / 2

  float bx, by;   // Ball (X, Y)
  int   bd;       // Ball (Direction, Spin)

  float sine[360], cosine[360];      // Generate trig tables
  int degree;
  for (degree = 0; degree < 360; degree++)
  {
    float radian = (float)degree * (M_PI / 180.0);
    sine[degree]   = sin(radian);
    cosine[degree] = cos(radian);
  }

#define P_START_H ((200 - 32 - PADDLE_HEIGHT) / 2)

  int p1, p2;

#define P1_X 10
#define P2_X 319 - 10 - 4

#define KEY_W  17
#define KEY_X  45
#define KEY_P  25

#define KEY_Q  16
#define KEY_Y  21
#define KEY_N  49

  float old_bx, old_by;

  int player1score = 0, player2score = 0;
  while (player1score < play_to && player2score < play_to)
  {
    bx = BALL_START_X;      // Ball coords
    by = BALL_START_Y;

    do
      { bd = random(360); }    // Ball direction
    while (!(bd <= 45 || (bd >= 135 && bd <= 135 + 90) || bd >= 315));

    p1 = P_START_H;  // Paddle heights
    p2 = P_START_H;

    DrawScreen();
    blitImage(P1_X, p1 + 8, &paddle1); // Draw first paddle
    blitImage(P2_X, p2 + 8, &paddle2); // Draw second paddle
    blitImage(bx - 1, by + 7, &ball);           // Draws new one

    clearArea(0, 183, 320, 16);
    blitString(0, 183, 15, "Press any key to start");
    char ss[41];
    sprintf(ss, "%-15d          %15d", player1score, player2score);
    blitString(0, 191, 15, ss);
    while (keyPressed())  {}
    while (!keyPressed()) {}
    clearArea(0, 183, 320, 8);

    int done = 0;
    while (!done)
    {
      if (keyTable[KEY_W])     if (p1 > 0)                        p1--;
      if (keyTable[KEY_X])     if (p1 < 200 - 32 - PADDLE_HEIGHT) p1++;

      if (num_players == 2)
      {
        if (keyTable[UPARROW])   if (p2 > 0)                        p2--;
        if (keyTable[DOWNARROW]) if (p2 < 200 - 32 - PADDLE_HEIGHT) p2++;
      }
      else
      {
        if (by < p2 + PADDLE_HEIGHT / 4     && random(2))
          if (p2 > 0)                        p2--;
        if (by > p2 + PADDLE_HEIGHT * 3 / 4 && random(2))
          if (p2 < 200 - 32 - PADDLE_HEIGHT) p2++;
      }

      if (keyTable[KEY_P])
      {
        blitString(0, 183, 15, "Press any key to continue");
        while (keyPressed())  {}
        while (!keyPressed()) {}
        clearArea(0, 183, 320, 8);
      }
      if (keyTable[KEY_Q])
      {
        blitString(0, 183, 15, "Quit: Are you sure? (Y/N)");

        int d = 0;
        while (!d)
        {
          if (keyTable[KEY_Y]) goto endofgame;
          if (keyTable[KEY_N]) d = 1;
        }

        clearArea(0, 183, 320, 8);
      }

      old_bx = bx; old_by = by;
      bx += cosine[bd] * 2;       // Move ball
      by += sine[bd]   * 2;
      if (by < 3)              // If ball hit top wall...
      {
        PlaySound(SOUND_HITWALL);
        bd = 360 - bd;
        bx = old_bx;
      }
      if (by > 200 - 32 - 3)   // If ball hit bottom wall...
      {
        PlaySound(SOUND_HITWALL);
        bd = 360 - bd;
        by = old_by;
      }

      if (bx + 1 < P1_X + 9)
      {
        if (by + 1 >= p1 - 1 && by + 1 <= p1 + PADDLE_HEIGHT) // Ball hit first paddle
        {
          PlaySound(SOUND_PADDLE);
          bd = 180 - bd;
          bx = old_bx; by = old_by;

          // Change ball direction a bit
          bd += random(21) - 10;
          if (bd < 0) bd += 360;
          else if (bd > 359) bd -= 360;
        }
      }

      if (bx + 1 > P2_X - 1)
      {
        if (by + 1 >= p2 - 1 && by + 1 <= p2 + PADDLE_HEIGHT) // Ball hit second paddle
        {
          PlaySound(SOUND_PADDLE);
          bd = 180 - bd;
          bx = old_bx; by = old_by;

          // Change ball direction a bit
          bd += random(21) - 10;
          if (bd < 0) bd += 360;
          else if (bd > 359) bd -= 360;
        }
      }

      if (p1 > 0)                        clearArea(P1_X, p1 + 7,                 4, 1, 1);
      if (p1 < 200 - 32 - PADDLE_HEIGHT) clearArea(P1_X, p1 + 8 + PADDLE_HEIGHT, 4, 1, 1);
      blitImage(P1_X, p1 + 8, &paddle1); // Draw first paddle
      if (p2 > 0) clearArea(P2_X, p2 + 7, 4, 1, 1);
      if (p2 < 200 - 32 - PADDLE_HEIGHT) clearArea(P2_X, p2 + 8 + PADDLE_HEIGHT, 4, 1, 1);
      blitImage(P2_X, p2 + 8, &paddle2); // Draw second paddle
      clearArea(old_bx - 1, old_by + 7, 3, 3, 1); // Erases previous ball
      blitImage(bx - 1, by + 7, &ball);           // Draws new one

      HDelay((11 - game_speed) << 10);

      if (bx + 1 < P1_X + 4)    // First player lost
      {
        PlaySound(SOUND_LOST);
        player2score++;
        done = 1;
      }

      if (bx + 1 > P2_X)        // Second player lost
      {
        PlaySound(SOUND_LOST);
        player1score++;
        done = 1;
      }
    }
  }

endofgame:

  resetImage(&paddle1);
  resetImage(&paddle2);
  resetImage(&ball);

  fadeOut(game_pal);
  loadScreen("tempscrn.tmp");
}

void DrawScreen(void)
{
  clearScreen(1);                     // Dark blue
  clearArea(0, 200 - 16, 320, 16, 0); // Text area

  int i;
  for (i = 0; i < 8; i++)             // Top and bottom borders
  {
    clearArea(0, i,            320, 1, i + 32);
    clearArea(0, 200 - 24 + i, 320, 1, 39 - i);
  }
}

void GenerateGamePal(void)
{
  memcpy(game_pal, default_pal, 256 * 3);

  // Leave first 16 colors alone  (standard)
  // Leave next 16 colors alone   (greyscale)

  int i;
  colorType clr;

  for (i = 32; i < 40; i++)      // Generates yellow gradient of 8 colors
  {
    clr.red   = (i - 32) * 8;
    clr.green = (i - 32) * 8;
    clr.blue  = 0;
    game_pal[i] = clr;
  }

  for (i = 40; i < 48; i++)      // Generates red gradient of 8 colors
  {
    clr.red   = (i - 40) * 8;
    clr.green = 0;
    clr.blue  = 0;
    game_pal[i] = clr;
  }

  for (i = 48; i < 56; i++)      // Generates green gradient of 8 colors
  {
    clr.red   = 0;
    clr.green = (i - 48) * 8;
    clr.blue  = 0;
    game_pal[i] = clr;
  }

  game_pal[56].red   = 63;       // Generates a white color
  game_pal[56].green = 63;
  game_pal[56].blue  = 63;

  game_pal[57].red   = 31;       // Generates a half-white, half-blue color
  game_pal[57].green = 31;
  game_pal[57].blue  = 63;
}

void HDelay(long msecs)
{
  HTimer tt;
  tt.timerOn();
  while (tt.getElapsed() < msecs) {}
  tt.timerOff();
}
